function [ximin, P, Measures]  = minXi(N, n, nu, degree)
%% Computes the minimax xi; prepares data for Table 3
%% N must be odd, for symmetry
%% Example:
% [ximin, P, Measures]  = minXi(51,20,.5,3)
rng(9)
chi = linspace(-1,1,N)'; %design space  
if degree==1 
    F = [ones(N,1), chi]; % Model matrix (X=JF); linear regression
elseif degree==2
    F = [ones(N,1), chi, chi.^2]; % Model matrix (X=JF); qudratic regression
elseif degree==3
    F = [ones(N,1), chi, chi.^2, chi.^3]; % Model matrix (X=JF); cubic regression
end
p = size(F,2);
[Q,~] = qr(F, 'econ'); % Orthogonal basis for F

fun = @(xi)xiloss(xi, n, Q, nu, p);
% Using particle swarm optimization:
fmincomopts = optimoptions('fmincon',  'algorithm', 'interior-point', 'OptimalityTolerance', 1e-12);
options = optimoptions('particleswarm', 'Display', 'iter', 'HybridFcn', {@fmincon, fmincomopts});
[ximin, ~] = particleswarm(fun, N, zeros(1,N), ones(1,N), options);

% Using a genetic algorithm (this is MUCH slower):
%options = optimoptions('ga', 'Display', 'iter', 'HybridFcn', @fmincon);
%[ximin, ~] = ga(fun, N, [],[],[],[],zeros(1,N), ones(1,N));

[ximin, ~] = roundXi(ximin,n);
[P, Measures] = summaryloss(ximin, n, Q, nu, p);
Measures = round(Measures,4);
nximin = n*ximin;

figure(4)
hold on
bar(chi, nximin, BarWidth=.1)
ylim([0, max(nximin)]);
yticks(1:max(nximin));
str2 = ['N = ', num2str(N), ', n = ', num2str(n), ', \nu = ',  num2str(nu), '.'];
if (degree == 1)
    str1 = 'Minimax design frequencies for straight line regression:';
    xlabel({str1; str2})
elseif (degree == 2)
    str1 = 'Minimax design frequencies for quadratic regression:';
    xlabel({str1; str2})
elseif (degree == 3)
    str1 = 'Minimax design frequencies for cubic regression:';
    xlabel({str1; str2})
end
hold off

%% ============================= subfunction "xiloss" =============================
function loss = xiloss(xi, n, Q, nu, p)
xi = xi/sum(xi);
xi = (xi + flip(xi))/2;  %symmetrize
N = length(xi);
[~, nxi] = roundXi(xi,n);
cdf = [0, cumsum(nxi)];
J = zeros(n,N); % incidence matrix of design
for j = 1:N
    if cdf(j+1) > cdf(j)
        J((cdf(j)+1):cdf(j+1),j) = ones(cdf(j+1) - cdf(j),1);
    end
end
L = minroot(n, J, Q, nu, p); 
P = L*L';
lossP = Loss(P, J, Q, nu, p);
loss = lossP;

%% ============================= subfunction "summaryloss" =============================
function [P, Measures] = summaryloss(xi, n, Q, nu, p)
xi = xi/sum(xi);
xi = (xi + flip(xi))/2;  %symmetrize
N = length(xi);
[~, nxi] = roundXi(xi,n);
cdf = [0, cumsum(nxi)];
J = zeros(n,N); % incidence matrix of design
for j = 1:N
    if cdf(j+1) > cdf(j)
        J((cdf(j)+1):cdf(j+1),j) = ones(cdf(j+1) - cdf(j),1);
    end
end
L = minroot(n, J, Q, nu, p); 
P = L*L';
I = eye(n);

[lossP, varP, biasP] = Loss(P, J, Q, nu, p);
[lossI, varI, biasI] = Loss(I, J, Q, nu, p);

% Percent improvement of P\nu, relative to P = I (i.e. OLS):
lossDecreaseOverOLS  = 100*(lossI - lossP)/lossI;

% Percent increase in variance, using this P, relative to P=I (i.e. OLS):
varIncreaseOverOLS = 100*(varP - varI)/varI;

% Percent decrease in bias, using this P, relative to P=I (i.e. OLS):
biasDecreaseOverOLS = 100*(biasI - biasP)/biasI ;

Measures = [lossI, lossP, lossDecreaseOverOLS, varIncreaseOverOLS, biasDecreaseOverOLS];


%% ============================= subfunction "minroot" =============================
% Minimizes the loss over triangular square roots of the precision matrix P
function L = minroot(n, J, Q, nu, p) % l0 = vecs L, lower triangular
func = @(l)lossroot(l, n, J, Q, nu, p);
nonlcon = @(l) mycon(l,n);
options = optimset('Display', 'none', 'algorithm', 'interior-point');
v = vecs(eye(n),n);
[l0, ~, ~] = fmincon(func, v,[],[],[],[],[],[], nonlcon, options); 
L = unvec(l0,n);

%% ============================= subfunction "mycon" =============================
function [c,ceq] = mycon(l,n)
c = [];
ceq = l'*l - n; 

%% ============================= subfunction "lossroot" =============================
%loss as function of root of P = L'*L; l = vecs L
function loss = lossroot(l, n, J, Q, nu, p) % l is n(n+1)/2 by 1
L = unvec(l,n);
% L is lower triangular, vecs(L) = l
[loss, ~, ~] = Loss(L*L', J, Q, nu, p);

%% ============================= subfunction "loss" =============================
function [loss, varonly, biasonly] = Loss(P, J, Q, nu, p)
U = J'*P*J;
V = J'*P*P*J;
U0 = Q'*U*Q;
U1 = Q'*U*U*Q;
V0 = Q'*V*Q;
U0inv = U0\eye(p);
varonly = trace(U0inv*V0*U0inv');
biasonly = max(eig(U0inv*U1*U0inv'));
loss = (1-nu)*varonly + nu*biasonly;

%% ============================= subfunction "roundXi" =============================
function [xi, nxi] = roundXi(xi,n)
% Takes an N-point design and rounds it so give integer frequencies summing to n
N = length(xi); % N must be odd
xi = xi/sum(xi);
xi = (xi + flip(xi))/2;
nxi = round(n*xi);
if(sum(nxi) < n)
    toofew =  ceil(n*xi) - n*xi;
    [~,I] = sort(toofew, 'descend');
    indices = 1:(n - sum(nxi));
    nxi(I(indices)) = nxi(I(indices)) + 1;
elseif(sum(nxi) > n)
    toomany =  n*xi - floor(n*xi);
    [~,I] = sort(toomany, 'descend');
    indices = 1:(sum(nxi) - n);
    nxi(I(indices)) = nxi(I(indices)) - 1;
end
I = 1:N;
I1 = I(nxi > (nxi+flip(nxi))/2);
nxi(I1) = nxi(I1) - 1;
nxi((N+1)/2) = nxi((N+1)/2) + length(I1);
xi = nxi/n;

%% ============================= subfunction "unvec" =============================
function L = unvec(l,n)
L = zeros(n,n);
for j = 1:n
    jj = (j-1)*(n-(j-2)/2);
    L(:,j) = [zeros(j-1,1); l((jj+1):(jj + n-j+1))];
end
% L is lower triangular, vecs(L) = l

%% ============================= subfunction "vecs" =============================
%Forms a vector from the lower triangle of a matrix
function vec = vecs(mat, p)
vec = [];
for j = 1:p
	vec = [vec; mat(j:p,j)];
end
 

